import bz2
import gzip
import json
import lzma
import os
import time


class StreamReceiver:
    """
    Base class for stream receivers.

    This isn't an ABC because I can imagine cases for super() chaining and I
    think subclassing code would discover introduced implementation errors
    quickly.
    """

    def send(self, msg):
        """
        :param msg: Either the raw bytes or a json-decoded object.
        """
        pass

    def close(self):
        pass


OPEN_FUNCS = {'.bz2': bz2.open, '.xz': lzma.open, '.gz': gzip.open}


def flexible_open(file_path, append=True):
    """
    Opens an output, binary file with inferred compression (or lack thereof).

    :param file_path: the output file path
    :param append: if False overwrite an existing file otherwise appends to it.
    :return: the opened file
    """
    mode = ('a' if append else 'w') + 'b'

    open_f = OPEN_FUNCS.get(os.path.splitext(file_path.lower())[-1], open)

    return open_f(file_path, mode)


class FeedSerializer(StreamReceiver):
    """
    Write each message to a bz2 file, *syncronously*.
    """

    def __init__(self, output_file, append=True):
        """
        :param as_bz2: if true, saves the file with BZ2 copression.
            BZ2 is useful for Spark/Hadoop analytics as it's splittable and
            compression rate is high. However, it's CPU intensive, which could
            be a problem.
        """
        self._output_stream = flexible_open(output_file, append)

    def send(self, msg):
        if not isinstance(msg, bytes):
            msg = json.dumps(msg).encode()
        self._output_stream.write(msg + b"\n")

    def close(self):
        self._output_stream.close()


class PartitioningFeedSerializer(StreamReceiver):
    """
    A Feed Serializer that creates a new archive partition every n messages.
    """

    def __init__(self, n_per_file, path_tmpl, append=False):
        """
        :param n_per_file: Messages per file before next partition created
        :param path_tmpl: a string representing the output path with a
            placeholder for format that injects the creation time. E.g.
            "my_log.{}.jsonl".
        """
        self._msgs_read = 0
        self._n_per_file = n_per_file
        self._path_tmpl = path_tmpl
        self._append = append

        file_path = self._path_tmpl.format(time.time())
        self._serializer = FeedSerializer(file_path, self._append)

    def send(self, msg):
        self._serializer.send(msg)

        self._msgs_read += 1

        if self._msgs_read >= self._n_per_file:
            self._msgs_read = 0
            self._serializer.close()

            file_path = self._path_tmpl.format(time.time())
            self._serializer = FeedSerializer(file_path, self._append)


class CallableStreamHandler(StreamReceiver):
    """
    Stream receiver that wraps a callable (mostly for testing)
    """

    def __init__(self, underlying_func):
        """
        :param underlying_func: function to invoke for each message received
            by ``#send(msg)``
        """
        self._underlying_func = underlying_func

    def send(self, msg):
        return self._underlying_func(msg)



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def bidding_rules(*, currency=_ELIDE):
    """
    Retrieve the bidding rules for a specific or all currencies. The response
    will indicate the minimum and maximum CPE (cost-per-engagement) bids.

    :param currency: The type of a currency to filter results by, identified
        using ISO-4217. This is a three-letter string “USD” or “EUR”. Omit this
        parameter to retrieve all bidding rules. associated with the
        authenticating user. Type: string Example: USD
    """
    binding = {'currency': currency}
    url = 'https://ads-api.twitter.com/2/bidding_rules'
    return _TwitterRequest('GET',
                           url,
                           'ads:bidding_rules',
                           'get-bidding-rules',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/bidding_rules'] = 'https://dev.twitter.com/ads/reference/get/bidding_rules'



import time
import pytest
from datetime import datetime
from brittle_wit.rate_limit import RateLimit, TimingRateLimiter
from tests.helpers import mock_resp, async_patch


def test_from_ignorance():
    rate_limit = RateLimit.from_ignorance()
    assert not rate_limit.is_exhausted


def test_from_initial_limit():
    rate_limit = RateLimit.from_initial_limit(10)
    assert rate_limit._remaining == 10


def test_from_response():
    resp = mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                      'X-RATE-LIMIT-REMAINING': 99,
                      'X-RATE-LIMIT-RESET': 9999})
    rate_limit = RateLimit.from_response(resp)
    local_dt = datetime.fromtimestamp(9999).strftime("%H:%M:%S")
    expected = "RateLimit(limit=100, remaining=99, reset_time={})"
    expected = expected.format(local_dt)

    assert rate_limit.reset_time == 9999
    assert rate_limit.__repr__() == expected
    assert str(rate_limit) == expected


def test_is_pristine():
    resp = mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                      'X-RATE-LIMIT-REMAINING': 100,
                      'X-RATE-LIMIT-RESET': 9999})
    rate_limit = RateLimit.from_response(resp)
    assert rate_limit.is_pristine

    resp = mock_resp({'X-RATE-LIMIT-REMAINING': 99})
    rate_limit.update(resp)
    assert not rate_limit.is_pristine


def test_is_exhausted():
    rate_limit = RateLimit.from_ignorance()
    assert not rate_limit.is_exhausted

    resp = mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                      'X-RATE-LIMIT-REMAINING': 0,
                      'X-RATE-LIMIT-RESET': 9999})

    rate_limit.update(resp)
    assert rate_limit.is_exhausted


def test_bad_request_update():
    orig_resp = mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                           'X-RATE-LIMIT-REMAINING': 99,
                           'X-RATE-LIMIT-RESET': 9999})

    rate_limit = RateLimit.from_response(orig_resp)
    assert not rate_limit.is_exhausted
    rate_limit.update(mock_resp({}))


def test_mark_exhausted():
    orig_resp = mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                           'X-RATE-LIMIT-REMAINING': 99,
                           'X-RATE-LIMIT-RESET': 9999})

    rate_limit = RateLimit.from_ignorance()
    assert not rate_limit.is_exhausted
    rate_limit.mark_exhausted(orig_resp)
    assert rate_limit.is_exhausted


@pytest.mark.asyncio
async def test_comply_without_sleeping():
    rate_limit = RateLimit.from_ignorance()
    assert not rate_limit.is_exhausted
    await rate_limit.comply()
    assert rate_limit.is_exhausted

    rate_limit.update(mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                                 'X-RATE-LIMIT-REMAINING': 0,
                                 'X-RATE-LIMIT-RESET': time.time() + 0.1}))
    assert rate_limit.is_exhausted
    await rate_limit.comply()
    assert not rate_limit.is_exhausted


@pytest.mark.asyncio
async def test_comply_with_sleeping():
    rate_limit = RateLimit.from_ignorance()
    assert not rate_limit.is_exhausted

    await rate_limit.comply()
    assert rate_limit.is_exhausted

    rate_limit.update(mock_resp({'X-RATE-LIMIT-LIMIT': 100,
                                 'X-RATE-LIMIT-REMAINING': 0,
                                 'X-RATE-LIMIT-RESET': time.time() + 360}))
    assert rate_limit.is_exhausted

    with async_patch('asyncio.sleep') as f:
        await rate_limit.comply()
    time_slept = f.call_args[0][0]
    assert time_slept < 360 + 5
    assert time_slept > 360 - 5
    assert not rate_limit.is_exhausted



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def uploadSTATUS(command, media_id, media_key):
    """
    The

    :param command: Must be set to STATUS (case sensitive).

    :param media_id: The media_id returned from the INIT command. Either this
        or media_key needs to be provided.

    :param media_key: The media_key returned from the INIT command. Either this
        or media_id needs to be provided.
    """
    binding = {'command': command, 'media_id': media_id, 'media_key':
               media_key}
    url = 'https://upload.twitter.com/1.1/media/upload.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:media',
                           'get-media-upload-status',
                           binding)


def metadata_create():
    """
    This endpoint can be used to provide additional information about the
    uploaded
    """
    binding = {}
    url = 'https://upload.twitter.com/1.1/media/metadata/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:media',
                           'post-media-metadata-create',
                           binding)


def upload(media, media_data, *, additional_owners=_ELIDE,
           media_category=_ELIDE):
    """
    Use this endpoint to upload images to Twitter. It returns a

    :param media: The raw binary file content being uploaded. Cannot be used
        with media_data.

    :param media_data: The base64-encoded file content being uploaded. Cannot
        be used with media.

    :param additional_owners: A comma-separated list of user IDs to set as
        additional owners allowed to use the returned media_id in Tweets or
        Cards. Up to 100 additional owners may be specified.

    :param media_category: A string enum value which identifies a media usecase
        (e.g. tweet_image, tweet_gif, tweet_video). If provided, media_key will
        be shown in the response.
    """
    binding = {'media': media, 'media_data': media_data, 'additional_owners':
               additional_owners, 'media_category': media_category}
    url = 'https://upload.twitter.com/1.1/media/upload.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:media',
                           'post-media-upload',
                           binding)


def uploadAPPEND(command, media_id, media_key, media, media_data,
                 segment_index):
    """
    The

    :param command: Must be set to APPEND (case sensitive).

    :param media_id: The media_id returned from the INIT command. Either this
        or media_key needs to be provided.

    :param media_key: The media_key returned from the INIT command. Either this
        or media_id needs to be provided.

    :param media: The raw binary file content being uploaded. It must be <= 5
        MB, and cannot be used with media_data.

    :param media_data: The base64-encoded chunk of media file. It must be <= 5
        MB and cannot be used with media. Use raw binary (media parameter) when
        possible.

    :param segment_index: An ordered index of file chunk. It must be between
        0-999 inclusive. The first segment has index 0, second segment has
        index 1, and so on.
    """
    binding = {'command': command, 'media_id': media_id, 'media_key':
               media_key, 'media': media, 'media_data': media_data,
               'segment_index': segment_index}
    url = 'https://upload.twitter.com/1.1/media/upload.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:media',
                           'post-media-upload-append',
                           binding)


def uploadFINALIZE(command, media_id, media_key):
    """
    The

    :param command: Must be set to FINALIZE (case sensitive).

    :param media_id: The media_id returned from the INIT command. Either this
        or media_key needd to be provided

    :param media_key: The media_key returned from the INIT command. Either this
        or media_id needs to be provided.
    """
    binding = {'command': command, 'media_id': media_id, 'media_key':
               media_key}
    url = 'https://upload.twitter.com/1.1/media/upload.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:media',
                           'post-media-upload-finalize',
                           binding)


def uploadINIT(command, total_bytes, media_type, *, media_category=_ELIDE,
               additional_owners=_ELIDE):
    """
    The

    :param command: Must be set to INIT (case sensitive).

    :param total_bytes: The size of the media being uploaded in bytes.

    :param media_type: The MIME type of the media being uploaded.

    :param media_category: A string enum value which identifies a media usecase
        (e.g. tweet_image, tweet_gif, tweet_video, amplify_video). This
        identifier is used to enforce usecase specific constraints (e.g.
        filesize, video duration) and enable advanced features.

    :param additional_owners: A comma-separated list of user IDs to set as
        additional owners allowed to use the returned media_id in Tweets or
        Cards. Up to 100 additional owners may be specified.
    """
    binding = {'command': command, 'total_bytes': total_bytes, 'media_type':
               media_type, 'media_category': media_category,
               'additional_owners': additional_owners}
    url = 'https://upload.twitter.com/1.1/media/upload.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:media',
                           'post-media-upload-init',
                           binding)


_TwitterRequest.DOC_URLS['https://upload.twitter.com/1.1/media/upload.json'] = 'https://dev.twitter.com/rest/reference/get/media/upload-status'
_TwitterRequest.DOC_URLS['https://upload.twitter.com/1.1/media/metadata/create.json'] = 'https://dev.twitter.com/rest/reference/post/media/metadata/create'
_TwitterRequest.DOC_URLS['https://upload.twitter.com/1.1/media/upload.json'] = 'https://dev.twitter.com/rest/reference/post/media/upload'
_TwitterRequest.DOC_URLS['https://upload.twitter.com/1.1/media/upload.json'] = 'https://dev.twitter.com/rest/reference/post/media/upload-append'
_TwitterRequest.DOC_URLS['https://upload.twitter.com/1.1/media/upload.json'] = 'https://dev.twitter.com/rest/reference/post/media/upload-finalize'
_TwitterRequest.DOC_URLS['https://upload.twitter.com/1.1/media/upload.json'] = 'https://dev.twitter.com/rest/reference/post/media/upload-init'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def accounts_by_account_id(account_id, end_time, entity, entity_ids,
                           granularity, metric_groups, placement, start_time):
    """
    Retrieve synchronous analytics for the current account.

    :param account_id: The identifier for the leveraged account. Appears within
        the resource’s path and is generally a required parameter for all
        Advertiser API requests excluding GET accounts. The specified account
        must be associated with the authenticating user. Type: string Example:
        18ce54d4x5t

    :param end_time: Scopes the retrieved data to the specified end time,
        expressed in ISO 8601. Note: Must be expressed in whole hours (0
        minutes and 0 seconds). Type: string Example: 2017-05-26T07:00:00Z

    :param entity: The entity type to retrieve data for. Type: enum Possible
        values: ACCOUNT, CAMPAIGN, FUNDING_INSTRUMENT, LINE_ITEM,
        ORGANIC_TWEET, PROMOTED_TWEET

    :param entity_ids: The specific entities to retrieve data for. Specify a
        comma-separated list of entity IDs. Note: Up to 20 entity IDs may be
        provided. Type: string Example: 8u94t

    :param granularity: Specify how granular the retrieved data should be.
        Type: enum Possible values: DAY, HOUR, TOTAL

    :param metric_groups: The specific metrics that should be returned. Specify
        a comma-separated list of metric groups. For more information see
        Metrics and Segmentation. Note: MOBILE_CONVERSION data should be
        requested separately. Type: enum Possible values: BILLING, ENGAGEMENT,
        LIFE_TIME_VALUE_MOBILE_CONVERSION, MEDIA, MOBILE_CONVERSION, VIDEO,
        WEB_CONVERSION

    :param placement: Scopes the retrieved data to a particular placement.
        Note: Only a single value accepted per request. For entities with both
        Twitter and Twitter Audience Platform placement, separate requests are
        required, one for each placement value. Type: enum Possible values:
        ALL_ON_TWITTER, PUBLISHER_NETWORK

    :param start_time: Scopes the retrieved data to the specified start time,
        expressed in ISO 8601. Note: Must be expressed in whole hours (0
        minutes and 0 seconds). Type: string Example: 2017-05-19T07:00:00Z
    """
    binding = {'account_id': account_id, 'end_time': end_time, 'entity':
               entity, 'entity_ids': entity_ids, 'granularity': granularity,
               'metric_groups': metric_groups, 'placement': placement,
               'start_time': start_time}
    url = 'https://ads-api.twitter.com/2/stats/accounts/{account_id}'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'ads:stats',
                           'get-stats-accounts-account-id',
                           binding)


def accounts_reach_campaigns_by_account_id(account_id, campaign_ids,
                                           end_time, start_time):
    """
    Retrieve reach and average frequency analytics for specified campaigns.

    :param account_id: The identifier for the leveraged account. Appears within
        the resource’s path and is generally a required parameter for all
        Advertiser API requests excluding GET accounts. The specified account
        must be associated with the authenticating user. Type: string Example:
        18ce54d4x5t

    :param campaign_ids: Scope the response to just the desired campaigns by
        specifying a comma-separated list of identifiers. Up to 20 IDs may be
        provided. Note: Up to 20 campaign IDs may be provided. Type: string
        Example: 8fgzf

    :param end_time: Scopes the retrieved data to the specified end time,
        expressed in ISO 8601. Note: Must be expressed in whole hours (0
        minutes and 0 seconds). Type: string Example: 2017-05-26T07:00:00Z

    :param start_time: Scopes the retrieved data to the specified start time,
        expressed in ISO 8601. Note: Must be expressed in whole hours (0
        minutes and 0 seconds). Type: string Example: 2017-05-19T07:00:00Z
    """
    binding = {'account_id': account_id, 'campaign_ids': campaign_ids,
               'end_time': end_time, 'start_time': start_time}
    url = 'https://ads-api.twitter.com/stats/accounts/{account_id}/reach/campaigns'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'ads:stats',
                           'get-stats-accounts-account-id-reach-campaigns',
                           binding)


def accounts_reach_funding_instruments_by_account_id(account_id,
                                                     funding_instrument_ids,
                                                     end_time, start_time):
    """
    Retrieve reach and average frequency analytics for specified funding
    instruments.

    :param account_id: The identifier for the leveraged account. Appears within
        the resource’s path and is generally a required parameter for all
        Advertiser API requests excluding GET accounts. The specified account
        must be associated with the authenticating user. Type: string Example:
        18ce54d4x5t

    :param funding_instrument_ids: Scope the response to just the desired
        funding instruments by specifying a comma-separated list of
        identifiers. Up to 20 IDs may be provided. Note: Up to 20 funding
        instrument IDs may be provided. Type: string Example: lygyi

    :param end_time: Scopes the retrieved data to the specified end time,
        expressed in ISO 8601. Note: Must be expressed in whole hours (0
        minutes and 0 seconds). Type: string Example: 2017-05-26T07:00:00Z

    :param start_time: Scopes the retrieved data to the specified start time,
        expressed in ISO 8601. Note: Must be expressed in whole hours (0
        minutes and 0 seconds). Type: string Example: 2017-05-19T07:00:00Z
    """
    binding = {'account_id': account_id, 'funding_instrument_ids':
               funding_instrument_ids, 'end_time': end_time, 'start_time':
               start_time}
    url = 'https://ads-api.twitter.com/stats/accounts/{account_id}/reach/funding_instruments'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'ads:stats',
                           'get-stats-accounts-account-id-reach-funding-instruments',
                           binding)


def jobs_accounts_by_account_id(account_id, *, count=_ELIDE, cursor=_ELIDE,
                                job_ids=_ELIDE):
    """
    Retrieve details for some or all asynchronous analytics job for the current
    account.

    :param account_id: The identifier for the leveraged account. Appears within
        the resource’s path and is generally a required parameter for all
        Advertiser API requests excluding GET accounts. The specified account
        must be associated with the authenticating user. Type: string Example:
        18ce54d4x5t

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example: 8x7v00oow

    :param job_ids: Scope the response to just the desired jobs by specifying a
        comma-separated list of identifiers. Up to 200 IDs may be provided.
        Type: long Example: 883787505404747776
    """
    binding = {'account_id': account_id, 'count': count, 'cursor': cursor,
               'job_ids': job_ids}
    url = 'https://ads-api.twitter.com/2/stats/jobs/accounts/{account_id}'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'ads:stats',
                           'get-stats-jobs-accounts-account-id',
                           binding)


def jobs_summaries(*, count=_ELIDE, cursor=_ELIDE):
    """
    Retrieve a summary of each existing asynchronous analytics job associated
    with the client app ID making the request. This endpoint is meant for
    internal / debugging purposes only.

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example:
        885662364120399872_4503599658438113
    """
    binding = {'count': count, 'cursor': cursor}
    url = 'https://ads-api.twitter.com/2/stats/jobs/summaries'
    return _TwitterRequest('GET',
                           url,
                           'ads:stats',
                           'get-stats-jobs-summaries',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/stats/accounts/{account_id}'] = 'https://dev.twitter.com/ads/reference/get/stats/accounts/account_id'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/stats/accounts/{account_id}/reach/campaigns'] = 'https://dev.twitter.com/ads/reference/get/stats/accounts/account_id/reach/campaigns'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/stats/accounts/{account_id}/reach/funding_instruments'] = 'https://dev.twitter.com/ads/reference/get/stats/accounts/account_id/reach/funding_instruments'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/stats/jobs/accounts/{account_id}'] = 'https://dev.twitter.com/ads/reference/get/stats/jobs/accounts/account_id'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/stats/jobs/summaries'] = 'https://dev.twitter.com/ads/reference/get/stats/jobs/summaries'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def available():
    """
    Returns the locations that Twitter has trending topic information for.
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/trends/available.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:trends',
                           'get-trends-available',
                           binding)


def closest(lat, long):
    """
    Returns the locations that Twitter has trending topic information for,
    closest to a specified location.

    :param lat: If provided with a long parameter the available trend locations
        will be sorted by distance, nearest to furthest, to the co-ordinate
        pair. The valid ranges for longitude is -180.0 to +180.0 (West is
        negative, East is positive) inclusive.

    :param long: If provided with a lat parameter the available trend locations
        will be sorted by distance, nearest to furthest, to the co-ordinate
        pair. The valid ranges for longitude is -180.0 to +180.0 (West is
        negative, East is positive) inclusive.
    """
    binding = {'lat': lat, 'long': long}
    url = 'https://api.twitter.com/1.1/trends/closest.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:trends',
                           'get-trends-closest',
                           binding)


def place(id, *, exclude=_ELIDE):
    """
    Returns the top 50 trending topics for a specific

    :param id: The Yahoo! Where On Earth ID of the location to return trending
        information for. Global information is available by using 1 as the
        WOEID.

    :param exclude: Setting this equal to hashtags will remove all hashtags
        from the trends list.
    """
    binding = {'id': id, 'exclude': exclude}
    url = 'https://api.twitter.com/1.1/trends/place.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:trends',
                           'get-trends-place',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/trends/available.json'] = 'https://dev.twitter.com/rest/reference/get/trends/available'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/trends/closest.json'] = 'https://dev.twitter.com/rest/reference/get/trends/closest'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/trends/place.json'] = 'https://dev.twitter.com/rest/reference/get/trends/place'



import pytest
import time
import asyncio
import random
from brittle_wit.rate_limit import RateLimit
from brittle_wit_core import (TwitterRequest, TwitterError,
                              AppCredentials, ClientCredentials)
from brittle_wit.executors import ManagedClientRequestProcessors, ANY_CREDENTIALS
from tests.helpers import async_ignore_sleep


@pytest.fixture
def alice():
    return ClientCredentials(1, "alice", "secret")


@pytest.fixture
def bob():
    return ClientCredentials(2, "bob", "secret")


@pytest.fixture
def manager():
    return ManagedClientRequestProcessors()


def test_nothing_managed_on_init(manager):
    assert not manager.is_managed(alice)


def test_managed_after_explicit_add(manager, alice):
    manager.add(alice)
    assert manager.is_managed(alice)


def test_managed_after_explicit_add_non_promiscuous(manager, alice):
    manager.add(alice, any_cred_eligible=False)
    assert manager.is_managed(alice)
    assert alice not in manager._any_cred_queue


def test_add_all(manager, alice, bob):
    manager.add_all([alice, bob])
    assert manager.is_managed(alice)
    assert manager.is_managed(bob)


def test_remove(manager, alice, bob):
    creds = [alice, bob]
    manager.add_all(creds)
    random.shuffle(creds)
    removed, kept = creds
    manager.remove(removed)
    assert kept in manager._any_cred_queue
    assert removed not in manager._any_cred_queue


def test_remove_not_any_cred(manager, alice):
    manager.add(alice, any_cred_eligible=False)
    assert alice not in manager._any_cred_queue
    manager.remove(alice)
    assert alice not in manager._any_cred_queue


def test_add_cannot_add_dup(manager, bob):
    manager.add(bob)
    with pytest.raises(RuntimeError):
        manager.add(bob)


def test__getitem__(manager, alice, bob):
    assert manager[alice]._client_cred == alice
    assert manager[bob]._client_cred == bob

    observed = {manager[ANY_CREDENTIALS]._client_cred for _ in range(10)}
    expected = {alice, bob}
    assert observed == expected


def test__getitem__anycred_exhausted(manager, bob):
    assert manager[ANY_CREDENTIALS] is None
    manager.add(bob)
    assert manager[ANY_CREDENTIALS] == bob


def test__maintain(manager, alice, bob):
    manager.add_all([alice, bob], False)

    now = time.time()
    manager[alice]._rate_limits['future'] = RateLimit(1, 1, now + 100)
    manager[bob]._rate_limits['passed'] = RateLimit(1, 1, now - 100)

    assert manager.is_managed(alice)
    assert manager.is_managed(bob)

    manager._maintain()

    assert manager.is_managed(alice)
    assert not manager.is_managed(bob)


def test__maintain_with_any_cred(manager, alice):
    manager.add_all([alice], True)
    now = time.time()
    manager[alice]._rate_limits['passed'] = RateLimit(1, 1, now - 100)
    assert manager.is_managed(alice)

    manager._maintain()
    assert manager.is_managed(alice)


@pytest.mark.asyncio
async def test_maintain_reraise_keyboard_interrupt(manager, mocker):
    path = 'brittle_wit.executors.ManagedClientRequestProcessors._maintain'
    _maintain = mocker.patch(path, side_effect=KeyboardInterrupt())
    with async_ignore_sleep() as sleep_times:
        with pytest.raises(KeyboardInterrupt):
            await manager.maintain()


@pytest.mark.asyncio
async def test_maintain_graceful_shutdown(manager, mocker):
    path = 'brittle_wit.executors.ManagedClientRequestProcessors._maintain'
    _maintain = mocker.patch(path, side_effect=asyncio.CancelledError())
    with async_ignore_sleep() as sleep_times:
        await manager.maintain()


@pytest.mark.asyncio
async def test_maintain_catches_all_else(manager, mocker):
    exceptions = [RuntimeError("test"), KeyboardInterrupt()]

    def f():
        raise exceptions.pop(0)

    path = 'brittle_wit.executors.ManagedClientRequestProcessors._maintain'
    _maintain = mocker.patch(path, side_effect=f)
    with async_ignore_sleep() as sleep_times:
        with pytest.raises(KeyboardInterrupt):
            await manager.maintain()



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def user(*, delimited=_ELIDE, stall_warnings=_ELIDE, with_=_ELIDE,
         replies=_ELIDE, track=_ELIDE, locations=_ELIDE,
         stringify_friend_ids=_ELIDE):
    """
    Streams messages for a single user, as described in

    :param delimited: Specifies whether messages should be length-delimited.
        See delimited for more information.

    :param stall_warnings: Specifies whether stall warnings should be
        delivered. See stall_warnings for more information.

    :param with_: Specifies whether to return information for just the
        authenticating user, or include messages from accounts the user
        follows. See with the with parameter documentation for more
        information.

    :param replies: Specifies whether to return additional @replies. See
        replies for more information.

    :param track: Includes additional Tweets matching the specified keywords.
        Phrases of keywords are specified by a comma-separated list. See track
        the track parameter documentation for more information.

    :param locations: Includes additional Tweets falling within the specified
        bounding boxes. See locations for more information.

    :param stringify_friend_ids: Specifies whether to send the Friend List
        preamble as an array of integers or an array of strings. See
        stringify_friends_id for more information.
    """
    binding = {'delimited': delimited, 'stall_warnings': stall_warnings,
               'with': with_, 'replies': replies, 'track': track, 'locations':
               locations, 'stringify_friend_ids': stringify_friend_ids}
    url = 'https://userstream.twitter.com/1.1/user.json'
    return _TwitterRequest('GET',
                           url,
                           'streaming:user',
                           'get-user',
                           binding)


_TwitterRequest.DOC_URLS['https://userstream.twitter.com/1.1/user.json'] = 'https://dev.twitter.com/streaming/reference/get/user'



import pytest
from brittle_wit.streaming import basic_streamer
from brittle_wit_core import AppCredentials, ClientCredentials, TwitterRequest
from tests.helpers import mock_stream, async_ignore_sleep, MockHTTPResp
import brittle_wit.streaming_api


app_cred = AppCredentials("app", "secret")
client_cred = ClientCredentials(1, "client", "secret")


@pytest.mark.asyncio
async def test_basic_streamer(mocker):
    twitter_req = brittle_wit.streaming_api.statuses.sample()
    callback = mocker.Mock()
    expected = [b'{"status": "a"}',
                b'{"status": "b"}',
                RuntimeError('Sentinel')]
    stream = mock_stream(expected)

    mocker.patch('brittle_wit.streaming.TwitterStream', return_value=stream)

    with pytest.raises(RuntimeError) as e:
        await basic_streamer(app_cred, client_cred, twitter_req, callback)
    assert 'Sentinel' in str(e)
    callbacks = [call[0][0] for call in callback.call_args_list]
    assert [{'status': "a"}, {'status': "b"}] == callbacks



from brittle_wit_core import AppCredentials, ClientCredentials, TwitterRequest
from brittle_wit.executors import debug_blocking_request


def test_debug_blocking_request(mocker):
    app_cred = AppCredentials("key", "secret")
    client_cred = ClientCredentials(1, "token-alice", "****")
    twitter_req = TwitterRequest("method", "url", "family", "service")

    f = mocker.patch('requests.request', return_value=42)
    res = debug_blocking_request(app_cred, client_cred, twitter_req)
    assert res == 42



import aiohttp
import asyncio
import bz2
import json
import time

from aiohttp import web
from brittle_wit.rate_limit import RateLimit
from brittle_wit.helpers import LOGGER
from brittle_wit_core import TwitterError, BrittleWitError
from brittle_wit.executors import twitter_req_to_http_req, LOGGER
import brittle_wit.connection_props as DEFAULTS


from brittle_wit.streaming.entry_processor import (EntryProcessor,
                                                   noop_entry_parser,
                                                   json_entry_parser)
from brittle_wit.streaming.receivers import (StreamReceiver,
                                             FeedSerializer,
                                             CallableStreamHandler,
                                             PartitioningFeedSerializer)
from brittle_wit.streaming.twitter_stream import TwitterStream



class StreamingHTTPPipe(StreamReceiver):
    """
    Utility StreamReceiver which pipes messages to streaming HTTP clients.

    Example
    -------
    > pipe = StreamingHTTPPipe()
    > stream_processor.subscribe(pipe, False)
    > app = aiohttp.web.Application()
    > app.router.add_get('/', pipe.handle())
    > server = loop.create_server(httpd.make_handler(), "0.0.0.0", 8394)
    > loop.run_until_complete(server)
    """
    DEFAULT_HEADERS = {'Content-Type': 'application/json'}

    def __init__(self, **resp_headers):
        # The _clients dict maps a unique identifier -- via id(obj) -- to a
        # pair of (StreamResponse, Event).
        self._clients = {}

        self._headers = {**self.DEFAULT_HEADERS, **resp_headers}

    async def handle(self, req):
        caller_id = id(req)
        resp = web.StreamResponse(status=200, headers=self._headers)

        try:
            # Start the FSM.
            await resp.prepare(req)

            # The send method signals handle when its processing finishes.
            finished_flag = asyncio.Event()
            self._clients[caller_id] = (resp, finished_flag)
            await finished_flag.wait()

        except asyncio.CancelledError:
            # Generally, this means the client disconnected.
            if caller_id in self._clients:
                del self._clients[caller_id]

        return resp

    def send(self, message):
        removal_set = []

        for caller_id, (resp, finished_flag) in self._clients.items():
            try:
                resp.write(message)  # I don't think draining is nessessary.
            except Exception as e:
                # This is a sloppy catch all for now. I'm not sure if
                # asyncio guarantees a cancel to handle prior to a
                # possible write. But, I do know it would be quite
                # wrong for a write to one client to cause failures
                # in other clients.
                finished_flag.set()
                removal_set.append(caller_id)

        for k in removal_set:
            del self._clients[k]


class StreamProcessor:
    """
    Sends TwitterResponses from a Stream to subscribers.
    """

    def __init__(self, twitter_stream):
        self._twitter_stream = twitter_stream
        self._subscribers = {}
        self._requires_json = False
        self._n_failures = False

    @property
    def requires_json(self):
        return self._requires_json

    def subscribe(self, receiver, as_json=False):
        """
        Subscribe a given receiver to the stream.

        :param receiver: a StreamReceiver object
        :param as_json: json parsing is relatively expensive. if a subscriber
            requires json, json parsing occurs, but only once for all the
            subscribers requiring json. Otherwise, there is no json parsing
        """
        # It's preferable to derive a class from StreamReceiver, but sometimes
        # it adds complexity. Send is the duck-type-required method.
        if not hasattr(receiver, "send"):
            raise AttributeError("Handler must implement send(msg)")

        handler_id = id(receiver)

        if handler_id in self._subscribers:
            raise RuntimeError("Already subscribed!")

        if as_json:
            self._requires_json = True

        self._subscribers[handler_id] = (receiver, as_json)
        return handler_id

    def unsubscribe(self, receiver):
        """
        Unsubscribe the given receiver from stream processing.

        :param receiver: an already subscribed receiver
        """
        handler_id = id(receiver)

        if handler_id not in self._subscribers:
            msg = "Either never subscribed or already unsubscribed!"
            raise RuntimeError(msg)

        required_json = self._subscribers[handler_id][1]

        del self._subscribers[handler_id]

        if required_json and not any(p[1] for p in self._subscribers.values()):
            self._requires_json = False

    def _send_to_all(self, message):
        # JSON parsing is expensive. If the application only pipes the
        # Twitter Stream to HTTP clients, it's dead time. The pattern
        # would be: BYTES -> JSON -> BYTES
        if self._requires_json:
            json_message = json_entry_parser(message)

        for subscriber, as_json in self._subscribers.values():
            subscriber.send(json_message if as_json else message)

    async def _stream_to_subscribers(self):
        LOGGER.info("Taking one for reset failures.")

        # Take one. If it's successful, this lets us reset the failure count.
        # One assignment for each send is a waste of cycles!
        async for message in self._twitter_stream:
            self._send_to_all(message)
            self._n_failures = 0
            break

        LOGGER.info("Streaming to subscribers.")

        async for message in self._twitter_stream:
            self._send_to_all(message)

    async def run(self):
        """
        Run the StreamProcessor, parsing the stream and sending responses.

        Retry Strategy: https://dev.twitter.com/streaming/overview/connecting
        """
        sleep_time = 0

        while True:
            LOGGER.info("Running StreamProcessor.")
            try:
                await self._stream_to_subscribers()
            except KeyboardInterrupt:
                LOGGER.error("User requested termination (KeyboardInterrupt)!")
                raise  # You're debugging, and want things to stop.
            except TwitterError as e:
                if e.status_code == 420:
                    # > Back off exponentially for HTTP 420 errors.
                    # > Start with a 1 minute wait and double each attempt.
                    sleep_time = min(2**self._n_failures, 16) * 60
                elif e.status_code == 429:
                    # You've been rate limited.
                    rate_limit = RateLimit.from_response(self._resp)
                    sleep_time = rate_limit.sleep_time_remaining
                else:
                    # > Back off exponentially for HTTP errors for which
                    # > reconnecting would be appropriate. Start with a 5
                    # > second wait, doubling each attempt, up to 320 seconds.
                    sleep_time = min(2**self._n_failures * 5, 320)

                msg = "Stream %s'd waiting a %d seconds (failure #%d)"
                LOGGER.error(msg, e, sleep_time, self._n_failures)
            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                # > These problems are generally temporary and tend to clear
                # > quickly. Increase the delay in reconnects by 250ms each
                # > attempt, up to 16 seconds.
                sleep_time = min(0.25 * (self._n_failures + 1), 16.0)

                msg = "Stream %s'd waiting a %d seconds (failure #%d)"
                LOGGER.error(msg, e, sleep_time, self._n_failures)
            except Exception as e:
                LOGGER.error("Aborting on %s", e)
                raise

            # You have to completely disconnect!
            self._twitter_stream.disconnect()
            self._n_failures += 1
            await asyncio.sleep(sleep_time)


async def basic_streamer(app_cred, client_cred, twitter_req, callback,
                         as_json=True):
    """
    Process a simple stream with a callback.

    :param app_cred: the AppCredentials
    :param client_cred: the ClientCredentials
    :param twitter_req: the TwitterRequest to initiate streaming
    :param callback: a callback in the form of ``f(tweet_obj)``
    :param as_json: if True (default) parse each message as json
    """
    handler = CallableStreamHandler(callback)

    opts = DEFAULTS.SESSION_OPTS.copy()
    opts['connector'] = aiohttp.TCPConnector(**DEFAULTS.TCP_CONN_STREAMING)

    with aiohttp.ClientSession(**opts) as sess:
        stream = TwitterStream(sess, app_cred, client_cred, twitter_req)
        processor = StreamProcessor(stream)
        processor.subscribe(handler, as_json)

        await processor.run()



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def app_store_categories(*, q=_ELIDE, store=_ELIDE):
    """
    Discover available app store category-based targeting criteria for Promoted
    Products. App store categories are available for the iOS App Store and the
    Google Play store only.

    :param q: An optional query to scope a targeting criteria. Omit this
        parameter to retrive all. Type: string Example: music

    :param store: Scope the results by a specific app store. Type: enum
        Possible values: GOOGLE_PLAY, IOS_APP_STORE
    """
    binding = {'q': q, 'store': store}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/app_store_categories'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-app-store-categories',
                           binding)


def behavior_taxonomies(*, behavior_taxonomy_ids=_ELIDE, count=_ELIDE,
                        cursor=_ELIDE, parent_behavior_taxonomy_ids=_ELIDE,
                        sort_by=_ELIDE, with_total_count=_ELIDE):
    """
    This endpoint is used to list the full (or partial) structure of behavior
    taxonomy trees.

    :param behavior_taxonomy_ids: Scope the response to just the desired
        behavior taxonomies by specifying a comma-separated list of
        identifiers. Up to 200 IDs may be provided. Type: string Example: 5e

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example: 8x7v00oow

    :param parent_behavior_taxonomy_ids: Scope the response to just the
        behavior taxonomies under specific parent nodes by specifying a comma-
        separated list of identifiers. Up to 200 IDs may be provided Type:
        string Example: 2e

    :param sort_by: Sorts by supported attribute in ascending or descending
        order. See Sorting for more information. Type: string Example:
        created_at-asc

    :param with_total_count: Include the total_count response attribute. Note:
        This parameter will be ignored if cursor is specified. Note: Requests
        which include total_count will have lower rate limits, currently set at
        200 per 15 minutes. Type: boolean Default: false Possible values: true,
        false
    """
    binding = {'behavior_taxonomy_ids': behavior_taxonomy_ids, 'count': count,
               'cursor': cursor, 'parent_behavior_taxonomy_ids':
               parent_behavior_taxonomy_ids, 'sort_by': sort_by,
               'with_total_count': with_total_count}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/behavior_taxonomies'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-behavior-taxonomies',
                           binding)


def behaviors(*, behavior_ids=_ELIDE, count=_ELIDE, country_code=_ELIDE,
              cursor=_ELIDE, sort_by=_ELIDE, with_total_count=_ELIDE):
    """
    Retrieves all valid behaviors defined by partner audiences that can be
    targeted by an advertiser.

    :param behavior_ids: Scope the response to just the desired behaviors by
        specifying a comma-separated list of identifiers. Up to 200 IDs may be
        provided. Type: string Example: lfrt

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param country_code: An optional query to scope a targeting criteria search
        to a specific country with the 2 letter ISO country code. If this
        parameter is not specified only partner audiences for the United States
        are returned. Type: string Default: US

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example: 8x7v00oow

    :param sort_by: Sorts by supported attribute in ascending or descending
        order. See Sorting for more information. Type: string Example:
        created_at-asc

    :param with_total_count: Include the total_count response attribute. Note:
        This parameter will be ignored if cursor is specified. Note: Requests
        which include total_count will have lower rate limits, currently set at
        200 per 15 minutes. Type: boolean Default: false Possible values: true,
        false
    """
    binding = {'behavior_ids': behavior_ids, 'count': count, 'country_code':
               country_code, 'cursor': cursor, 'sort_by': sort_by,
               'with_total_count': with_total_count}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/behaviors'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-behaviors',
                           binding)


def devices(*, count=_ELIDE, q=_ELIDE):
    """
    Discover available device-based targeting criteria for Promoted Products.
    Device targeting is available for Promoted Tweets.

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param q: An optional query to scope a targeting criteria. Omit this
        parameter to retrive all. Type: string Example: apple
    """
    binding = {'count': count, 'q': q}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/devices'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-devices',
                           binding)


def events(event_types, *, count=_ELIDE, country_code=_ELIDE, cursor=_ELIDE,
           end_time=_ELIDE, ids=_ELIDE, start_time=_ELIDE):
    """
    Discover available event-based targeting criteria for Promoted Products.
    Only one event can be targeted per line item.

    :param event_types: An optional query to scope to certain event types.
        Type: enum Possible values: CONFERENCE, HOLIDAY,
        MUSIC_AND_ENTERTAINMENT, OTHER, SPORTS

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param country_code: An optional query to scope a targeting criteria search
        to a specific country with the 2 letter ISO country code. If this
        parameter is not specified only partner audiences for the United States
        are returned. Type: string Default: US

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example: 8x7v00oow

    :param end_time: The time, expressed in ISO 8601, that the campaign will
        end. Type: string Example: 2017-10-05T00:00:00Z

    :param ids: Scope the response to just the desired events by specifying a
        comma-separated list of identifiers. Up to 200 IDs may be provided.
        Type: string Example: lex

    :param start_time: The time, expressed in ISO 8601, that the line item will
        begin serving. Note: Defaults to the current time. Type: string
        Example: 2017-07-05T00:00:00Z
    """
    binding = {'event_types': event_types, 'count': count, 'country_code':
               country_code, 'cursor': cursor, 'end_time': end_time, 'ids':
               ids, 'start_time': start_time}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/events'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-events',
                           binding)


def interests():
    """
    Discover available interest-based targeting criteria for Promoted Products.
    Interests change infrequently, however we suggest you refresh this list at
    least once weekly.
    """
    binding = {}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/interests'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-interests',
                           binding)


def languages(*, count=_ELIDE, cursor=_ELIDE, q=_ELIDE):
    """
    Discover languages available for targeting.

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example: 8x7v00oow

    :param q: An optional query to scope a targeting criteria. Omit this
        parameter to retrive all. Type: string Example: english
    """
    binding = {'count': count, 'cursor': cursor, 'q': q}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/languages'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-languages',
                           binding)


def locations(*, count=_ELIDE, country_code=_ELIDE, cursor=_ELIDE,
              location_type=_ELIDE, q=_ELIDE):
    """
    Discover available location-based targeting criteria for Promoted Products.
    Geo-targeting is available for Promoted Accounts and Promoted Tweets at the
    country level, state/region level, city level, and postal code level.
    Postal code targeting must be used if you wish to retrieve analytics at the
    postal code level.

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param country_code: An optional query to scope a targeting criteria search
        to a specific country with the 2 letter ISO country code. Omit this
        parameter to retrieve results for all countries. Type: string Example:
        JP

    :param cursor: Specifies a cursor to get the next page of results. See
        Pagination for more information. Type: string Example: 8x7v00oow

    :param location_type: Scope the results by a specific kind of location.
        More granular targeting than COUNTRIES may not be available in all
        locations. Type: enum Possible values: COUNTRIES, REGIONS, METROS,
        CITIES, POSTAL_CODES

    :param q: An optional query to scope a targeting criteria search. Omit this
        parameter to retrieve all results. Type: string Example: New York
    """
    binding = {'count': count, 'country_code': country_code, 'cursor': cursor,
               'location_type': location_type, 'q': q}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/locations'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-locations',
                           binding)


def network_operators(*, country_code=_ELIDE, q=_ELIDE, count=_ELIDE,
                      cursor=_ELIDE):
    """
    Target users based on mobile carrier (AT&T, Verizon, Sprint, T-Mobile, etc)
    in multiple countries. This endpoint enables you to lookup targetable
    carriers available for targeting with the

    :param country_code: An optional parameter that specifies the country_code
        to query for from available network operators.

    :param q: An optional query to scope a search. Omit this parameter to
        retrieve all results.

    :param count: Specifies the number of locations to try and retrieve, up to
        a maximum of 1000 per distinct request.

    :param cursor: Specifies a cursor to get the next page of locations. See
        Pagination for more information.
    """
    binding = {'country_code': country_code, 'q': q, 'count': count, 'cursor':
               cursor}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/network_operators'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-network-operators',
                           binding)


def platform_versions(*, q=_ELIDE):
    """
    Discover available mobile-OS-version-based targeting criteria for Promoted
    Products. Platform version targeting is available for Promoted Accounts and
    Promoted Tweets. This allows targeting down to the point release of a
    mobile operating system version, for example iOS 7.1, Android 4.4

    :param q: An optional query to scope a targeting criteria search. Omit this
        parameter to retrieve all results.
    """
    binding = {'q': q}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/platform_versions'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-platform-versions',
                           binding)


def platforms(*, count=_ELIDE, q=_ELIDE, lang=_ELIDE):
    """
    Discover available platform-based targeting criteria for Promoted Products.

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param q: An optional query to scope a targeting criteria search. Omit this
        parameter to this parameter to retrieve all results. Type: string
        Examples: ios, blackberry

    :param lang: Using a ISO-639-1 language code. When passed, an additional
        localized_name attribute will be returned in the response. Type: int,
        string Example: fr
    """
    binding = {'count': count, 'q': q, 'lang': lang}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/platforms'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-platforms',
                           binding)


def tv_markets():
    """
    Discover available TV markets where TV shows can be targeted. Returns
    markets by locale that can used to query the
    """
    binding = {}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/tv_markets'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-tv-markets',
                           binding)


def tv_shows(tv_market_locale, *, q=_ELIDE, count=_ELIDE, cursor=_ELIDE):
    """
    Discover available TV show targeting criteria for Promoted Products. TV
    show targeting is available for Promoted Tweets in certain markets. See

    :param tv_market_locale: A required parameter that specifies the
        tv_market_locale to query for available TV shows. TV markets are
        queried based on locale returned from the GET
        targeting_criteria/tv_markets endpoint.

    :param q: An optional query to scope a targeting criteria search. Omit this
        parameter to retrieve all results.

    :param count: Specifies the number of TV shows to try and retrieve, up to a
        maximum of 50 per distinct request.

    :param cursor: Specifies a cursor to get the next page of results. See
        pagination for more information.
    """
    binding = {'tv_market_locale': tv_market_locale, 'q': q, 'count': count,
               'cursor': cursor}
    url = 'https://ads-api.twitter.com/2/targeting_criteria/tv_shows'
    return _TwitterRequest('GET',
                           url,
                           'ads:targeting_criteria',
                           'get-targeting-criteria-tv-shows',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/app_store_categories'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/app_store_categories'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/behavior_taxonomies'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/behavior_taxonomies'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/behaviors'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/behaviors'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/devices'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/devices'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/events'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/events'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/interests'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/interests'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/languages'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/languages'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/locations'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/locations'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/network_operators'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/network_operators'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/platform_versions'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/platform_versions'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/platforms'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/platforms'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/tv_markets'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/tv_markets'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/targeting_criteria/tv_shows'] = 'https://dev.twitter.com/ads/reference/get/targeting_criteria/tv_shows'



__title__ = "brittle_wit"
__description__ = "Brittle Wit is a Twitter Lib for Python."
__uri__ = "https://github.com/jbn/brittle_wit"
__doc__ = __description__ + " <" + __uri__ + ">"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2017 John Bjorn Nelson"
__version__ = "0.0.5"
__author__ = "John Bjorn Nelson"
__email__ = "jbn@abreka.com"


from brittle_wit_core import AppCredentials, ClientCredentials
from brittle_wit_core import BrittleWitError, Cursor, ELIDE
from brittle_wit.executors import (ANY_CREDENTIALS,
                                   ClientRequestProcessor,
                                   LOGGER,
                                   ManagedClientRequestProcessors)
from brittle_wit.streaming import (basic_streamer,
                                   StreamProcessor,
                                   StreamingHTTPPipe,
                                   TwitterStream)
from brittle_wit.app import App
from brittle_wit.helpers import monkey_patch_ipython_disp


def _patch_retryables():
    import aiohttp
    import asyncio
    import brittle_wit_core
    exceptions = brittle_wit_core.WrappedException.RETRYABLE_EXCEPTIONS
    exceptions.add(aiohttp.ClientOSError)
    exceptions.add(aiohttp.ClientConnectorError)
    exceptions.add(asyncio.TimeoutError)

_patch_retryables()



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def list(*, user_id=_ELIDE, screen_name=_ELIDE, count=_ELIDE,
         since_id=_ELIDE, max_id=_ELIDE, include_entities=_ELIDE):
    """
    Returns the 20 most recent Tweets favorited by the authenticating or
    specified user.

    :param user_id: The ID of the user for whom to return results for.

    :param screen_name: The screen name of the user for whom to return results
        for.

    :param count: Specifies the number of records to retrieve. Must be less
        than or equal to 200; defaults to 20. The value of count is best
        thought of as a limit to the number of tweets to return because
        suspended or deleted content is removed after the count has been
        applied.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param include_entities: The entities node will be omitted when set to
        false.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'count': count,
               'since_id': since_id, 'max_id': max_id, 'include_entities':
               include_entities}
    url = 'https://api.twitter.com/1.1/favorites/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:favorites',
                           'get-favorites-list',
                           binding)


def create(id, *, include_entities=_ELIDE):
    """
    Favorites the status specified in the ID parameter as the authenticating
    user. Returns the favorite status when successful.

    :param id: The numerical ID of the desired status.

    :param include_entities: The entities node will be omitted when set to
        false.
    """
    binding = {'id': id, 'include_entities': include_entities}
    url = 'https://api.twitter.com/1.1/favorites/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:favorites',
                           'post-favorites-create',
                           binding)


def destroy(id, *, include_entities=_ELIDE):
    """
    Un-favorites the status specified in the ID parameter as the authenticating
    user. Returns the un-favorited status in the requested format when
    successful.

    :param id: The numerical ID of the desired status.

    :param include_entities: The entities node will be omitted when set to
        false.
    """
    binding = {'id': id, 'include_entities': include_entities}
    url = 'https://api.twitter.com/1.1/favorites/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:favorites',
                           'post-favorites-destroy',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/favorites/list.json'] = 'https://dev.twitter.com/rest/reference/get/favorites/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/favorites/create.json'] = 'https://dev.twitter.com/rest/reference/post/favorites/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/favorites/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/favorites/destroy'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def filter(*, follow=_ELIDE, track=_ELIDE, locations=_ELIDE,
           delimited=_ELIDE, stall_warnings=_ELIDE):
    """
    Returns public statuses that match one or more filter predicates. Multiple
    parameters may be specified which allows most clients to use a single
    connection to the Streaming API. Both GET and POST requests are supported,
    but GET requests with too many parameters may cause the request to be
    rejected for excessive URL length. Use a POST request to avoid long URLs.

    :param follow: A comma separated list of user IDs, indicating the users to
        return statuses for in the stream. See follow for more information.

    :param track: Keywords to track. Phrases of keywords are specified by a
        comma-separated list. See track for more information.

    :param locations: Specifies a set of bounding boxes to track. See locations
        for more information.

    :param delimited: Specifies whether messages should be length-delimited.
        See delimited for more information.

    :param stall_warnings: Specifies whether stall warnings should be
        delivered. See stall_warnings for more information.
    """
    binding = {'follow': follow, 'track': track, 'locations': locations,
               'delimited': delimited, 'stall_warnings': stall_warnings}
    url = 'https://stream.twitter.com/1.1/statuses/filter.json'
    return _TwitterRequest('POST',
                           url,
                           'streaming:statuses',
                           'post-statuses-filter',
                           binding)


def sample(*, delimited=_ELIDE, stall_warnings=_ELIDE):
    """
    Returns a small random sample of all public statuses. The Tweets returned
    by the default access level are the same, so if two different clients
    connect to this endpoint, they will see the same Tweets.

    :param delimited: Specifies whether messages should be length-delimited.
        See delimited or more information.

    :param stall_warnings: Specifies whether stall warnings should be
        delivered. See stall_warnings for more information.
    """
    binding = {'delimited': delimited, 'stall_warnings': stall_warnings}
    url = 'https://stream.twitter.com/1.1/statuses/sample.json'
    return _TwitterRequest('GET',
                           url,
                           'streaming:statuses',
                           'get-statuses-sample',
                           binding)


_TwitterRequest.DOC_URLS['https://stream.twitter.com/1.1/statuses/filter.json'] = 'https://dev.twitter.com/streaming/reference/post/statuses/filter'
_TwitterRequest.DOC_URLS['https://stream.twitter.com/1.1/statuses/sample.json'] = 'https://dev.twitter.com/streaming/reference/get/statuses/sample'



import json
from collections import deque


class EntryProcessor:
    """
    Splits a stream of bytes into CRLF-separated messages.
    """

    def __init__(self):
        # Note: I tried using a bytearray instead of a byte string but it
        # ended up being slower and having greater jitter than byte strings.
        self._buf = b""
        self._mailbox = deque()

    def process(self, chunk):
        """
        Translate received bytes into messages in a mailbox.

        :param chunk: a chunk of bytes received from the HTTP connection
        """
        self._buf += chunk

        # Streams are `\r\n`-separated JSON messages.
        raw_lines = self._buf.split(b"\r\n")

        # If only one element in the split, then there wasn't a CRLF.
        if len(raw_lines) > 1:

            # The last element may be a b'', which is perfectly fine.
            self._buf = raw_lines[-1]

            # Blank lines are keep-alive messages.
            self._mailbox.extend(l for l in raw_lines[:-1] if l.strip())

    def messages(self):
        """
        Return the messages then reset the mailbox

        :return: the enqueued messages as a list
        """
        return list(iter(self))

    def __bool__(self):
        """
        :return: True if there are any messages available
        """
        return bool(self._mailbox)

    def pop(self):
        """
        :return: the next message in the mailbox
        """
        return self._mailbox.popleft()

    def purge_buffer(self):
        """
        Reset the internal buffer.

        This is useful when restarting a stream. When the old stream
        terminates, the buffer is incomplete. Appending to it will produce a
        corrupt entry.
        """
        self._buf = b''

    def purge_mailbox(self):
        """
        Resets the mailbox so there are no messages
        """
        self._mailbox.clear()

    def __iter__(self):
        """
        :return: an iterator which pops messages of the mailbox until fully
            consumed
        """
        return self._entry_consumer()

    def _entry_consumer(self):
        while self._mailbox:
            yield self.pop()


def noop_entry_parser(entry):
    """
    :return: the entry string without further processing.
    """
    return entry


def json_entry_parser(entry):
    """
    :return: the entry parsed as a JSON object
    """
    return json.loads(entry.decode('ascii'))



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def list():
    """
    Returns the authenticated user’s saved search queries.
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/saved_searches/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:saved_searches',
                           'get-saved-searches-list',
                           binding)


def show_by_id(id):
    """
    Retrieve the information for the saved search represented by the given id.
    The authenticating user must be the owner of saved search ID being
    requested.

    :param id: The ID of the saved search.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/saved_searches/show/{id}.json'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'rest:saved_searches',
                           'get-saved-searches-show-id',
                           binding)


def create(query):
    """
    Create a new saved search for the authenticated user. A user may only have
    25 saved searches.

    :param query: The query of the search the user would like to save.
    """
    binding = {'query': query}
    url = 'https://api.twitter.com/1.1/saved_searches/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:saved_searches',
                           'post-saved-searches-create',
                           binding)


def destroy_by_id(id):
    """
    Destroys a saved search for the authenticating user. The authenticating
    user must be the owner of saved search id being destroyed.

    :param id: The ID of the saved search.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/saved_searches/destroy/{id}.json'
    url = url.format(**binding)
    return _TwitterRequest('POST',
                           url,
                           'rest:saved_searches',
                           'post-saved-searches-destroy-id',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/saved_searches/list.json'] = 'https://dev.twitter.com/rest/reference/get/saved_searches/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/saved_searches/show/{id}.json'] = 'https://dev.twitter.com/rest/reference/get/saved_searches/show/id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/saved_searches/create.json'] = 'https://dev.twitter.com/rest/reference/post/saved_searches/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/saved_searches/destroy/{id}.json'] = 'https://dev.twitter.com/rest/reference/post/saved_searches/destroy/id'
